(({ Toolshed, behaviors, debounce }) => {
  /**
   * Object which represents 1 autocomplete value to select. This object
   * manages its own resources and the display text as well as the underlying
   * value to pass back to the original input element.
   */
  class SuggestItem {
    /**
     * Create a new SuggestItem representing values and a selectable element.
     *
     * @param {SuggestList} list
     *   The suggestion list that this item is a member of.
     * @param {function(event)} clickHandler
     *   The callback when the item is clicked on.
     */
    constructor(list, clickHandler) {
      this.parent = list;

      this.pos = -1;
      this.el = document.createElement('A');
      this.el._suggestItem = this;
      this.el.tabIndex = '-1';
      this.el.className = 'autocomplete__suggestion';
      this.el.addEventListener('click', clickHandler);
    }

    /**
     * Fetch the root DOM element for suggestion containers to include.
     *
     * @return {DOMElement}
     *   The root DOM element for this suggester.
     */
    getElement() {
      return this.el;
    }

    /**
     * Get the SuggestList containing this SuggestItem.
     *
     * @return {SuggestList}
     *   The parent SuggestList for this item.
     */
    getParent() {
      return this.parent;
    }

    /**
     * Set the SuggestItem to belong to this parent SuggestList.
     *
     * @param {SuggestList} parent
     *   The SuggestList to set as the parent of this item.
     */
    setParent(parent) {
      this.parent = parent;
    }

    /**
     * Set the display text for the select option.
     *
     * @param {string} text
     *   The displayed HTML and text for the autocomplete input.
     */
    setLabel(text) {
      this.el.innerHTML = text;
    }

    /**
     * Set the values for this suggestion item.
     *
     * @param {string} key
     *   The actual value to pass back as the result of the selection.
     */
    setValue(key) {
      this.value = key;
    }

    /**
     * Get the display and input value for this autocomplete item.
     *
     * @return {string|object}
     *   A JSON object with a key (the form submit value) and the
     *   text (display) value.
     */
    getValue() {
      return this.value;
    }

    /**
     * Update this item to show it has focus.
     */
    focus() {
      this.el.classList.add('autocomplete__suggestion--active');
    }

    /**
     * Update this item to show that it doesn't have focus.
     */
    blur() {
      this.el.classList.remove('autocomplete__suggestion--active');
    }

    /**
     * Clean up listeners and other allocated resources.
     */
    destroy() {
      this.el.removeEventListener('click', this.onClick);

      this.parent = null;
      delete this.el;
      delete this.wrapper;
    }
  }

  /**
   * A list of SuggestItems, and manage the interactions with the autocomplete
   * suggestions traversal, building and clearing.
   */
  class SuggestList {
    /**
     * Create a new list of SuggestItem objects.
     *
     * @param {SuggestList|null} list
     *   An object that contains a list of SuggestItem instances.
     * @param {function} clickHandler
     *   The on click callback to react to SuggestItem clicks.
     */
    constructor(list, clickHandler) {
      this.parent = list;
      this.pos = 0;
      this.items = [];
      this.onClick = clickHandler;

      this.label = document.createElement('SPAN');
      this.label.className = 'autocomplete__caption';

      this.el = document.createElement('UL');
      this.el.className = 'autocomplete__suggestions';

      this.wrap = document.createElement('DIV');
      this.wrap.appendChild(this.label);
      this.wrap.appendChild(this.el);
    }

    /**
     * Is this SuggestList empty of selectable suggestion items?
     *
     * @return {Boolean}
     *   Returns true if there are no selectable items currently available
     *   as suggestions for the autocomplete query.
     */
    isEmpty() {
      return !this.items.length;
    }

    /**
     * Get the DOM element which wraps a set of selectable suggests or lists.
     *
     * @return {DOMElement}
     *   The DOM element which wraps the selectable list of options.
     */
    getElement() {
      return this.wrap;
    }

    /**
     * Fetch the reference to the parent containing SuggestList.
     *
     * @return {SuggestList|null}
     *   The containing SuggestList for this list, or NULL if the root.
     */
    getParent() {
      return this.parent;
    }

    /**
     * Set the parent for the SuggestList.
     *
     * @param {SuggestList|null} parent
     *   The parent SuggestList to set as the parent of this list.
     */
    setParent(parent) {
      this.parent = parent;
    }

    /**
     * Set the label to display over a set items, if there is a caption.
     *
     * @param {string} text
     *   The label text to use as a caption over the SuggestList items.
     */
    setLabel(text) {
      this.label.textContent = (text && text.length) ? text : '';
    }

    /**
     * Get the first item in the list.
     *
     * @return {SuggestItem|null}
     *   The first selectable SuggestItem or NULL if not items.
     */
    getFirst() {
      if (this.items.length) {
        return this.items[0] instanceof SuggestList ? this.items[0].getFirst() : this.items[0];
      }

      return null;
    }

    /**
     * Get the list item in the list.
     *
     * @return {SuggestItem|null}
     *   The last selectable SuggestItem in the list or NULL. This includes
     *   the last item in the nested set of items.
     */
    getLast() {
      if (this.items.length) {
        const idx = this.items.length - 1;
        return this.items[idx] instanceof SuggestList ? this.items[idx].getLast() : this.items[idx];
      }

      return null;
    }

    /**
     * Construct the list of selectable SuggestItems from the passed in data.
     *
     * @param {array|object} data
     *   The data representing the autocomplete suggestions / values available.
     * @param {SuggestItem[]} pool
     *   A pool of available SuggestItem to reuse.
     */
    buildItems(data, pool = []) {
      let pos = 0;

      data.forEach((row) => {
        let item;

        if (row.list && Array.isArray(row.list)) {
          if (!row.list.length) return;

          item = new SuggestList(this, this.onClick);
          item.setLabel(row.text);
          item.buildItems(row.list);
        }
        else if (row.value) {
          item = (pool.length) ? pool.pop() : new SuggestItem(this, this.onClick);
          item.setLabel(row.text || row.value);
          item.setValue(row.value);
        }

        item.pos = pos++;
        item.setParent(this);
        this.items.push(item);

        const li = document.createElement('LI');
        li.appendChild(item.getElement());
        this.el.appendChild(li);
      });
    }

    /**
     * Remove suggestion items from the suggestioned items display.
     */
    clear() {
      this.items.forEach((item) => {
        try {
          const itemEl = item.getElement();
          const li = itemEl.parentNode;

          if (li.tagName === 'LI') {
            li.parentNode.removeChild(li);
          }

          li.removeChild(itemEl);
          item.destroy();
        }
        catch (e) {
          // Remove the element failed due to not being a child.
        }
      });

      this.items = [];
    }

    /**
     * Cleans up this autocomplete suggestions list, and frees resources and
     * event listeners.
     */
    destroy() {
      this.items.forEach((item) => item.destroy());

      this.wrap.parentNode.removeChild(this.wrap);

      delete this.el;
      delete this.wrap;
      delete this.label;
    }
  }

  /**
   * Class which encapsulates the behaviors of and autocomplete widget of a
   * textfield. It allows subclasses to override their fetchSuggestions() and
   * init() methods to allow for different methods of loading autocomplete
   * items to display.
   */
  Toolshed.Autocomplete = class {
    constructor(input, config = {}) {
      this.input = input;
      this.config = {
        delay: 375,
        minLength: 3,
        ...config,
      };

      this.pending = false;

      // Bind the function calls to this instance of the object.
      this.onTextKeydown = this.onTextKeydown.bind(this);
      this.onTextChange = this.onTextChange.bind(this);
      this.onParamChange = this.onParamChange.bind(this);
      this.onFocus = this.onFocus.bind(this);
      this.onBlur = this.onBlur.bind(this);
      this.onMouseDown = this.onMouseDown.bind(this);
      this.onItemClick = this.onItemClick.bind(this);

      if (this.config.delay > 0) {
        this.fetchSuggestions = debounce(this.fetchSuggestions, this.config.delay);
      }
    }

    /**
     * Initialize the autocomplete instance by attaching events and creating
     * autocomplete elements.
     */
    init() {
      if (this.isInit) return;

      this.isInit = true;

      const ac = this.input.cloneNode();
      ac.type = 'text';
      ac.classList.remove('toolshed-autocomplete');
      ac.classList.add('form-autocomplete');
      ac.placeholder = this.input.placeholder;
      ac.autocomplete = 'off';
      ac.value = this.input.dataset.text || this.input.value || '';

      // Remove these Drupal properties as this is just a stand-in object
      // and we don't want to submit or to catch any Drupal behaviors.
      delete ac.dataset.drupalSelector;
      delete ac.dataset.autocompletePath;
      delete ac.dataset.text;
      ac.removeAttribute('name');

      if (this.input.id) {
        ac.id = `${this.input.id}--autocomplete`;

        // Reassociate a label element to point to the new autocomplete input.
        const inputLabel = document.querySelector(`label[for='${this.input.id}']`);
        if (inputLabel) {
          inputLabel.htmlFor = ac.id;
        }
      }
      this.input.style.display = 'none';

      // Bind key change events.
      ac.addEventListener('keydown', this.onTextKeydown);
      ac.addEventListener('input', this.onTextChange);
      ac.addEventListener('focus', this.onFocus);
      ac.addEventListener('blur', this.onBlur);

      const wrapper = document.createElement('DIV');
      wrapper.style.display = 'none';
      wrapper.classList.add('autocomplete__suggestion-wrapper');
      wrapper.addEventListener('mousedown', this.onMouseDown);

      const emptyMsg = document.createElement('DIV');
      emptyMsg.style.display = 'none';
      emptyMsg.classList.add('autocomplete__empty-msg');
      wrapper.appendChild(emptyMsg);

      const list = new SuggestList(null, this.onItemClick);
      list.label.style.display = 'none';
      wrapper.appendChild(list.getElement());

      const { parentElement: parent, nextSibling } = this.input;
      parent.insertBefore(ac, nextSibling);
      parent.insertBefore(wrapper, nextSibling);

      this.ac = ac;
      this.list = list;
      this.emptyMsg = emptyMsg;
      this.suggestWrap = wrapper;

      if (this.config.params && this.input.form) {
        const { form } = this.input;

        Object.values(this.config.params).forEach((elemId) => {
          const el = form.querySelector(`#${elemId}`);

          if (el) el.addEventListener('change', this.onParamChange);
        });
      }
    }

    /**
     * Refresh the autocomplete suggests to display.
     *
     * @param {Array|Object} [data={}]
     *   Values to use as the autocomplete suggestions. The property keys are
     *   the value, and the property values are the display autocomplete value.
     */
    createSuggestions(data = {}) {
      this.activeItem = null;

      if (!this.list.isEmpty()) {
        this.list.clear();
      }

      if (data.list && data.list.length) {
        if (this.emptyMsg.style.display !== 'none') {
          this.emptyMsg.style.display = 'none';
        }

        this.list.buildItems(data.list);
      }
      else {
        this.emptyMsg.style.display = '';
        this.emptyMsg.textContent = (data.empty && data.empty.length) ? data.empty : 'No results';
      }
    }

    /**
     * Clear the current input.
     */
    clearInput() {
      this.ac.value = '';
      this.input.value = '';
      this.activeItem = null;
      this.clearSuggestions();
    }

    /**
     * Remove existing autocomplete suggestions.
     */
    clearSuggestions() {
      this.activeItem = null;
      this.list.clear();
    }

    /**
     * Is the suggestion window open?
     *
     * @return {Boolean}
     *   TRUE if the suggestions window is open, otherwise FALSE.
     */
    isSuggestionsVisible() {
      return this.suggestWrap.style.display !== 'none';
    }

    /**
     * Position and expose the suggestions window if it is not already open.
     */
    displaySuggestions() {
      if (!this.isSuggestionsVisible()) {
        this.suggestWrap.style.display = '';
        this.suggestWrap.style.width = `${this.ac.clientWidth}px`;
        this.suggestWrap.style.top = `${this.ac.offsetTop + this.ac.offsetHeight}px`;
        this.suggestWrap.style.left = `${this.ac.offsetLeft}px`;
      }
    }

    /**
     * Hide the suggestions window if it is open and reset the active item.
     */
    hideSuggestions() {
      this.activeItem = null;
      this.suggestWrap.style.display = 'none';
    }

    /**
     * Set a SuggestItem as currently active based on the index. This method
     * ensures that any currently active items are blurred (unfocused) and
     * if the requested index is out of range, to select no items.
     *
     * @param {SuggestItem} item
     *   The item to set as the current active item.
     */
    setActiveItem(item) {
      if (this.activeItem) this.activeItem.blur();

      if (item) {
        item.focus();
        this.activeItem = item;
      }
      else {
        this.activeItem = null;
      }
    }

    /**
     * Transfer the values from the passed in item to the form elements.
     *
     * @param {SuggestItem} item
     *   Item to apply to as the values of the autocomplete widget.
     */
    applyValue(item) {
      this.ac.value = item.getElement().textContent;
      this.input.value = item.getValue();
    }

    /**
     * Make the necessary requests and calls to get available text values.
     *
     * @param {string} text
     *   The text to try to match using the autocomplete service to
     *   generate suggestions with.
     * @param {bool} display
     *   Should the suggestions list be displayed after fetching suggestions.
     */
    fetchSuggestions(text, display = true) {
      if (!this.requester) {
        this.requester = Toolshed.createRequester(this.config.uri);
      }

      if (this.pending && !this.pending.promise.isResolved) {
        // Abort any previously pending requests, so this late response
        // won't overwrite our desired values if it returns out of order.
        this.pending.xhr.abort();
      }

      if (!text || text.length < this.config.minLength) {
        return;
      }

      // Turn on the loading spinner.
      this.ac.classList.add('ui-autocomplete-loading');

      // Apply additional request parameters if there are linked inputs.
      const requestParams = {};
      if (this.config.params) {
        Object.entries(this.config.params).forEach(([key, elemId]) => {
          const el = document.getElementById(elemId);

          if (el && el.value) requestParams[key] = el.value;
        });
      }
      requestParams.q = text;

      this.pending = this.requester(requestParams);
      this.pending.promise.then(
        (response) => {
          this.ac.classList.remove('ui-autocomplete-loading');
          this.createSuggestions(response);

          // If requested to display suggestions after they load.
          if (display) this.displaySuggestions();
        },
        (reason) => {
          // Only remove the autocomplete loading if the request was aborted.
          if (!(reason.message && reason.message === 'Cancelled')) {
            this.ac.classList.remove('ui-autocomplete-loading');
          }
        },
      );
    }

    /**
     * When focus is on the autocomplete input, show suggestions if they exist.
     */
    onFocus() {
      if (!this.list.isEmpty()) {
        this.displaySuggestions();
      }
    }

    /**
     * When focus leaves the autocomplete input, hide suggestions.
     */
    onBlur() {
      if (this.isSuggestionsVisible()) {
        this.hideSuggestions();
      }
    }

    /**
     * Create a mouse down event which avoids having the AC input lose focus
     * from clicking into the suggestions wrapper. This inconveniently prevents
     * the click event because it hides the suggestions before the click event
     * can occur.
     *
     * @param {Event} event
     *   The mouse down event object.
     */
    // eslint-disable-next-line class-methods-use-this
    onMouseDown(event) {
      event.preventDefault();
    }

    /**
     * Callback for updating the input value when the textfield is updated.
     *
     * @param {Event} event
     *   The input changed event object.
     */
    onTextChange(event) {
      this.input.value = '';
      this.fetchSuggestions(event.target.value);
    }

    onParamChange() {
      this.input.value = '';
      this.ac.value = '';
      this.clearSuggestions();
    }

    /**
     * Respond to keyboard navigation, and move the active item.
     *
     * @param {KeydownEvent} event
     *   Respond to a key-up event for the autocomplete input textfield.
     */
    onTextKeydown(event) {
      if (!this.isSuggestionsVisible()) return;

      // Escape key.
      if (event.keyCode === 27) {
        this.hideSuggestions();
      }

      // Enter key.
      if (event.keyCode === 13) {
        if (this.activeItem) {
          event.preventDefault();
          this.applyValue(this.activeItem);
        }

        this.hideSuggestions();
      }

      // Arrow keys (up and down).
      if (event.keyCode === 40 || event.keyCode === 38) {
        event.preventDefault();

        const [inc, method] = event.keyCode === 40
          ? [1, SuggestList.prototype.getFirst]
          : [-1, SuggestList.prototype.getLast];

        let item = this.activeItem;
        if (item) {
          let p = item.getParent();
          let i = item.pos + inc;

          while (p && (i < 0 || p.items.length === i)) {
            item = p;
            i = item.pos + inc;
            p = item.getParent();
          }

          if (p) {
            item = p.items[i];
          }
        }

        item = item || this.list;
        if (item instanceof SuggestList) {
          item = method.call(item);
        }

        this.setActiveItem(item);
      }
    }

    /**
     * A suggestion element was clicked. Updates the internal values.
     *
     * @param {ClickEvent} event
     *   The click event containing the context and the details of the
     *   triggered event.
     */
    onItemClick(event) {
      event.preventDefault();

      this.applyValue(event.target._suggestItem);
      this.hideSuggestions();
    }

    /**
     * Clean up DOM and event listeners initialized by this autocompete widget.
     */
    destroy() {
      if (this.pending && !this.pending.promise.isResolved) {
        // Abort any request for this widget, we no longer want the results.
        this.pending.xhr.abort();
      }
      delete this.pending;

      if (this.ac) {
        this.ac.removeEventListener('keydown', this.onTextKeydown);
        this.ac.removeEventListener('input', this.onTextChange);
        this.ac.removeEventListener('focus', this.onFocus);
        this.ac.removeEventListener('blur', this.onBlur);

        if (this.config.params) {
          Object.values(this.config.params).forEach((elemId) => {
            const el = document.getElementById(elemId);

            if (el) el.removeEventListener('change', this.onParamChange);
          });
        }

        if (this.ac.id && this.input.id) {
          // Reassociate a label element to point to the original input.
          const inputLabel = document.querySelector(`label[for='${this.ac.id}']`);
          if (inputLabel) {
            inputLabel.htmlFor = this.input.id;
          }
        }
        this.ac.parentElement.removeChild(this.ac);

        this.suggestWrap.removeEventListener('mousedown', this.onMouseDown);
        this.suggestWrap.parentElement.removeChild(this.suggestWrap);
        this.list.destroy();

        delete this.suggestWrap;
        delete this.emptyMsg;
        delete this.ac;
      }

      this.isInit = false;
      this.input.style.display = '';
    }
  };

  /**
   * Find autocomplete inputs and attach the autocomplete behaviors to it.
   */
  behaviors.toolshedAutocomplete = {
    // Track instances of Autocomplete objects so they can be detached.
    instances: new Map(),

    attach(context) {
      Toolshed.walkByClass(context, 'toolshed-autocomplete', (item) => {
        const settings = item.dataset.autocomplete ? JSON.parse(item.dataset.autocomplete) : {};

        if (item.dataset.params) {
          settings.params = JSON.parse(item.dataset.params);
        }

        if (settings.uri) {
          const ac = new Toolshed.Autocomplete(item, settings);

          this.instances.set(item.id || item, ac);
          ac.init();
        }
      }, 'autocomplete--processed');
    },

    detach(context, settings, trigger) {
      if (trigger === 'unload') {
        Toolshed.walkBySelector(context, '.toolshed-autocomplete.autocomplete--processed', (item) => {
          const ac = this.instances.get(item.id || item);

          if (ac) {
            item.classList.remove('autocomplete--processed');
            this.instances.delete(item);
            ac.destroy();
          }
        });
      }
    },
  };
})(Drupal);
