(({ Toolshed }) => {
  /**
   * A Pane is floating DIV with manageable content.
   *
   * It is designed to be the base of dialogs, pop-ups or other screen elements
   * that shows above the content that shows and hides outside the normal
   * document flow.
   */
  class Pane {
    static defaultOptions() {
      return {
        classes: [],
      };
    }

    /**
     * Create a new instance of the a display pane.
     *
     * @param {Object} options
     *   The configuration options for how to handle pane behaviors.
     */
    constructor(options = {}) {
      this.opts = {
        ...Pane.defaultOptions(),
        ...options,
      };
    }

    /**
     * Initialize the Pane by creating internal elements and attach event listeners.
     */
    init() {
      if (!this.el) {
        this.el = document.createElement('DIV');
        this.content = document.createElement('DIV');
        this.el.appendChild(this.content);

        this.applyStyles();

        this.append = !this.opts.onAddItem ? this.appendElement : (item) => {
          const el = this.opts.onAddItem(item);
          this.appendElement(el || item);
        };

        this.remove = !this.opts.onRemoveItem ? this.removeElement : (item) => {
          const el = this.opts.onRemoveItem(item);
          this.removeElement(el || item);
        };

        this.close();
        document.body.appendChild(this.el);
      }
    }

    /**
     * Apply the static styles to the pane based on the current options.
     */
    applyStyles() {
      this.el.style.position = 'absolute';

      this.el.classList.add('pane');
      this.el.classList.add('toolshed-pane');
      (this.opts.classes || []).forEach((className) => this.el.classList.add(className));

      this.content.classList.add('pane__content');

      // Apply Z-index if something was explicitly set for this pane.
      if (this.opts.zIndex) {
        this.el.style.zIndex = this.opts.zIndex;
      }
    }

    /**
     * Get the element which wraps the pane's content.
     *
     * @return {HTMLElement}
     *   The HTML element which wraps the content for this Pane instance.
     */
    getContent() {
      return this.content;
    }

    /**
     * Add an element to the content container.
     *
     * @param {Node} item
     *   The node to add to the content container.
     */
    appendElement(item) {
      if (item instanceof Node) {
        this.getContent().appendChild(item);
      }
    }

    /**
     * Remove an element from the content container.
     *
     * @param {Node} item
     *   The node to remove from the content container.
     */
    removeElement(item) {
      if (item instanceof Node) {
        this.getContent().removeChild(item);
      }
    }

    /**
     * Clear the content container of all nodes and elements.
     */
    clear() {
      const el = this.getContent();

      while (el.lastElementChild) {
        this.remove(el.lastElementChild);
      }
    }

    /**
     * Is the suggestion window open?
     *
     * @return {Boolean}
     *   TRUE if the suggestions window is open, otherwise FALSE.
     */
    isVisible() {
      return this.el.style.display !== 'none';
    }

    /**
     * Position the pane before displaying it.
     */
    // eslint-disable-next-line class-methods-use-this
    positionPane() {
      // TODO: position based on pane options.
    }

    /**
     * Position and expose the suggestions window if it is not already open.
     *
     * @return {bool}
     *   Always returns true, to indicate that the dialog is open.
     */
    open() {
      if (!this.isVisible()) {
        this.el.style.display = '';

        this.positionPane();
      }
      return true;
    }

    /**
     * Close the content pane.
     *
     * @return {bool}
     *   Always returns false, to indicate that the pane is closed.
     */
    close() {
      this.el.style.display = 'none';
      return false;
    }

    /**
     * Toggle to open and close status of the pane.
     *
     * @return {bool}
     *   Returns true if result is pane is open, and false if it is closed.
     */
    toggle() {
      return this.isVisible() ? this.close() : this.open();
    }

    /**
     * Teardown and clean-up of Pane contents.
     */
    destroy() {
      if (this.el) {
        this.clear();

        this.el.parentNode.removeChild(this.el);
        delete this.el;
      }
    }
  }

  /**
   * A pane, that attaches itself to another element in the document.
   *
   * AttachedPanes are a good base for pop-ups, toolbars or other elements
   * that display content and anchor themselves to other screen elements.
   */
  class AttachedPane extends Pane {
    /**
     * Create a new instance of an AttachedPane.
     *
     * @param {DOMElement} attachTo
     *   The element containing the pane content.
     * @param {Object} options
     *   The configuration options for how to handle pane behaviors.
     */
    constructor(attachTo, options = {}) {
      super(options);

      this.attachTo = attachTo;
    }

    /**
     * @override
     */
    positionPane() {
      const rect = this.attachTo.getBoundingClientRect();
      const scroll = {
        x: document.documentElement.scrollLeft || document.body.scrollLeft,
        y: document.documentElement.scrollTop || document.body.scrollTop,
      };

      this.el.style.top = (rect.top > this.el.clientHeight)
        ? `${(scroll.y - this.el.clientHeight) + rect.bottom}px`
        : `${scroll.y + rect.top}px`;

      this.el.style.left = `${scroll.x + rect.right}px`;

      // Calculate the z-index of the attached button, this becomes important,
      // when dealing with form elements in pop-up dialoges. Only calculate this
      // if not already explicitly set by the options.
      if (!this.opts.zIndex) {
        let node = this.attachTo;

        while (node) {
          const z = (node.style || {}).zIndex;

          if (Number.isInteger(z) || /\d+/.test(z)) {
            this.el.style.zIndex = z + 10;
            break;
          }

          node = node.parentNode;
        }
      }
    }
  }

  Toolshed.Pane = Pane;
  Toolshed.AttachedPane = AttachedPane;
})(Drupal);
