<?php

namespace Drupal\toolshed\Plugin;

use Drupal\Component\Plugin\PluginManagerBase;
use Drupal\Component\Plugin\Discovery\CachedDiscoveryInterface;
use Drupal\Component\Plugin\Discovery\DiscoveryCachedTrait;
use Drupal\Core\Plugin\Factory\ContainerFactory;
use Drupal\Core\Plugin\Discovery\ContainerDerivativeDiscoveryDecorator;
use Drupal\Core\Plugin\Discovery\YamlDiscovery;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Cache\UseCacheBackendTrait;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Extension\ThemeHandlerInterface;

/**
 * Plugin manager for plugins that are defined from YAML files.
 */
class YamlPluginManager extends PluginManagerBase implements CachedDiscoveryInterface {

  use DiscoveryCachedTrait;
  use UseCacheBackendTrait;

  /**
   * The name of the plugin to use as in the discovery.
   *
   * The string should contain only characters [a-z][0-9], periods or
   * underscores and will be used in the discovery handler to find definition
   * files with the naming convention "<module>.<pluginName>.yml".
   *
   * @var string
   */
  protected $pluginName;

  /**
   * The fully scoped class name, of the default plugin implementation class.
   *
   * @var string
   */
  protected $defaultPluginClass;

  /**
   * Module extension management service.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface|\Drupal\Core\Extension\ThemeHandlerInterface
   */
  protected $extensionHandler;

  /**
   * The caching backend for storing the plugin definitions.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  protected $cacheBackend;

  /**
   * The key of the cache data, holding the discovered plugin definitions.
   *
   * @var string
   */
  protected $cacheKey;

  /**
   * Create a new instance of the plugin manager using YAML file discovery.
   *
   * @param string $scope_name
   *   The name of the provider of the plugin manager. Typically this is the
   *   module name, where the manager is defined.
   * @param string $plugin_interface
   *   Interface which the plugin instances need to implement. Reinforces
   *   base functionality that utilizers of the plugin can assume.
   * @param string $default_class
   *   The default plugin class to use when the plugin definition does not
   *   include information about the class type.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface|\Drupal\Core\Extension\ThemeHandlerInterface $extension_handler
   *   Service which manages active modules.
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache_backend
   *   The cache to store the plugin discovery information.
   * @param string $cache_key
   *   The key to use when caching the discovery of plugin definitions.
   */
  public function __construct($scope_name, $plugin_interface, $default_class, $extension_handler, CacheBackendInterface $cache_backend = NULL, $cache_key = NULL) {
    if (!($extension_handler instanceof ModuleHandlerInterface || $extension_handler instanceof ThemeHandlerInterface)) {
      throw new \InvalidArgumentException();
    }

    $this->pluginName = $scope_name;
    $this->pluginInterface = $plugin_interface;
    $this->defaultPluginClass = $default_class;
    $this->extensionHandler = $extension_handler;

    if (isset($cache_backend)) {
      $this->cacheBackend = $cache_backend;
      $this->cacheKey = !empty($cache_key) ? $cache_key : strtr($scope_name, '.', '_');
    }
  }

  /**
   * {@inheritdoc}
   */
  protected function providerExists($provider) {
    // Check the if the provider extension exists.
    return ($this->extensionHandler instanceof ModuleHandlerInterface)
      ? $this->extensionHandler->moduleExists($provider)
      : $this->themeHandler->themeExists($provider);
  }

  /**
   * {@inheritdoc}
   */
  public function getDiscovery() {
    if (!$this->discovery) {
      $directories = ($this->extensionHandler instanceof ModuleHandlerInterface)
        ? $this->extensionHandler->getModuleDirectories()
        : $this->extensionHandler->getThemeDirectories();

      $discovery = new YamlDiscovery($this->pluginName, $directories);
      $discovery->addTranslatableProperty('label', 'label_context');
      $discovery = new ContainerDerivativeDiscoveryDecorator($discovery);
      $this->discovery = $discovery;
    }

    return $this->discovery;
  }

  /**
   * {@inheritdoc}
   */
  public function getFactory() {
    if (!$this->factory) {
      $this->factory = new ContainerFactory($this, $this->pluginInterface);
    }

    return $this->factory;
  }

  /**
   * {@inheritdoc}
   */
  public function useCaches($use_caches = FALSE) {
    $this->useCaches = $use_caches;

    if (!$this->useCaches) {
      $this->definitions = NULL;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function clearCachedDefinitions() {
    if ($this->cacheBackend) {
      $this->cacheBackend->delete($this->cacheKey);
    }

    $this->definitions = NULL;
  }

  /**
   * Use the plugin manager discovery to find and process definitions.
   *
   * @return array
   *   An array of plugins discovered and processed by they discovery instance.
   */
  protected function findDefinitions() {
    $definitions = $this->getDiscovery()
      ->getDefinitions();

    foreach ($definitions as $plugin_id => &$definition) {
      $this->processDefinition($definition, $plugin_id);
    }

    return $definitions;
  }

  /**
   * Process individual raw definitions from the discovery handler.
   *
   * Apply defaults and alter the raw definitions as needed to ensure that the
   * definitions are complete and allow for common defaults.
   *
   * @param mixed $definition
   *   Refereance to an array of plugin definition information that can be
   *   altered with defaults or other scope changes.
   * @param string $plugin_id
   *   The ID of the plugin being processed.
   */
  protected function processDefinition(&$definition, $plugin_id) {
    if (is_array($definition) && empty($definition['class'])) {
      $definition['class'] = $this->defaultPluginClass;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getDefinitions() {
    if (!isset($this->definitions)) {
      if ($cache = $this->cacheGet($this->cacheKey)) {
        $this->definitions = $cache->data;
      }
      else {
        $this->definitions = $this->findDefinitions();
        $this->cacheSet($this->cacheKey, $this->definitions);
      }
    }

    return $this->definitions;
  }

}
